#include "angelscript.h"
#include "scriptstdstring.h"

#include "as_smart_ptr_wrapper.h"
#include "as_gen_wrapper.h"

#include <iostream>
#include <cassert>
#include <cstdio>

#include <memory>

// Types to export

struct Base {
  virtual ~Base() {}

  void moo(void) const { std::cout << "Moo." << std::endl; }
};

struct Derived : Base {
  int x;
  int y;

  void moo(void) const { std::cout << "Moo!" << std::endl; }
  void add(int value) { x += value; y += value; }

  int & getter(void) { return x; }
 };

typedef std::shared_ptr<Base> BasePtr;
typedef std::shared_ptr<Derived> DerivedPtr;

// Helper functions

template <typename T>
void construct(void * address) {
  new (address) T;
}

template <typename T>
void destroy(T * object) {
  object->~T();
}

template <typename T>
void copy_construct(void * address, T * other) {
  new (address) T(*other);
}

template <typename T>
void assign(T * lhs, T* rhs) {
  *lhs = *rhs;
}

void MessageCallback(const asSMessageInfo *msg, void *param)
{
  const char *type = "ERR ";
  if (msg->type == asMSGTYPE_WARNING) {
    type = "WARN";
  } else if (msg->type == asMSGTYPE_INFORMATION) {
    type = "INFO";
  }

  printf("%s (%d, %d) : %s : %s\n", msg->section, msg->row, msg->col, type, msg->message);
}

void __cdecl print(const std::string & str) {
  std::cout << str;
}

template <typename T>
T foo(T);

int main(int, char **) {
  asIScriptEngine * engine = asCreateScriptEngine(ANGELSCRIPT_VERSION);
  engine->SetMessageCallback(asFUNCTION(MessageCallback), 0, asCALL_CDECL);
  RegisterStdString(engine);

  bool use_generic = strstr(asGetLibraryOptions(), "AS_MAX_PORTABILITY");

  int r;
  if (use_generic) {
    r = engine->RegisterGlobalFunction("void print(string & in)", WRAP_FN(print), asCALL_GENERIC); assert(r >= 0);
  } else {
    r = engine->RegisterGlobalFunction("void print(string & in)", asFUNCTION(print), asCALL_CDECL); assert(r >= 0);
  }

  // Base
  r = engine->RegisterObjectType("Base", sizeof(BasePtr), asOBJ_VALUE | asOBJ_APP_CLASS_CDAK); assert(r >= 0);
  if (use_generic) {
    r = engine->RegisterObjectBehaviour("Base", asBEHAVE_CONSTRUCT,  "void f()",                WRAP_OBJ_FIRST(construct<BasePtr>),      asCALL_GENERIC); assert(r >= 0);
    r = engine->RegisterObjectBehaviour("Base", asBEHAVE_CONSTRUCT,  "void f(const Base & in)", WRAP_OBJ_FIRST(copy_construct<BasePtr>), asCALL_GENERIC); assert(r >= 0);
    r = engine->RegisterObjectBehaviour("Base", asBEHAVE_DESTRUCT,   "void f()",                WRAP_OBJ_FIRST(destroy<BasePtr>),        asCALL_GENERIC); assert(r >= 0);
    r = engine->RegisterObjectMethod("Base", "Base &opAssign(const Base &in)", WRAP_OBJ_FIRST(assign<BasePtr>), asCALL_GENERIC); assert(r >= 0);

    r = engine->RegisterObjectMethod("Base", "void moo(void) const", GEN_CALLER(Base, moo), asCALL_GENERIC); assert(r >= 0);
  } else {
    r = engine->RegisterObjectBehaviour("Base", asBEHAVE_CONSTRUCT,  "void f()",                asFUNCTION(construct<BasePtr>),      asCALL_CDECL_OBJFIRST); assert(r >= 0);
    r = engine->RegisterObjectBehaviour("Base", asBEHAVE_CONSTRUCT,  "void f(const Base & in)", asFUNCTION(copy_construct<BasePtr>), asCALL_CDECL_OBJFIRST); assert(r >= 0);
    r = engine->RegisterObjectBehaviour("Base", asBEHAVE_DESTRUCT,   "void f()",                asFUNCTION(destroy<BasePtr>),        asCALL_CDECL_OBJFIRST); assert(r >= 0);
    r = engine->RegisterObjectMethod("Base", "Base &opAssign(const Base &in)", asFUNCTION(assign<BasePtr>), asCALL_CDECL_OBJFIRST); assert(r >= 0);

    r = engine->RegisterObjectMethod("Base", "void moo(void) const", CALLER(Base, moo), asCALL_CDECL_OBJFIRST); assert(r >= 0);
  }

  // Derived
  r = engine->RegisterObjectType("Derived", sizeof(DerivedPtr), asOBJ_VALUE | asOBJ_APP_CLASS_CDAK); assert(r >= 0);
  if (use_generic) {
    r = engine->RegisterObjectBehaviour("Derived", asBEHAVE_CONSTRUCT,  "void f()",                   WRAP_OBJ_FIRST(construct<DerivedPtr>),      asCALL_GENERIC); assert(r >= 0);
    r = engine->RegisterObjectBehaviour("Derived", asBEHAVE_CONSTRUCT,  "void f(const Derived & in)", WRAP_OBJ_FIRST(copy_construct<DerivedPtr>), asCALL_GENERIC); assert(r >= 0);
    r = engine->RegisterObjectBehaviour("Derived", asBEHAVE_DESTRUCT,   "void f()",                   WRAP_OBJ_FIRST(destroy<DerivedPtr>),        asCALL_GENERIC); assert(r >= 0);
    r = engine->RegisterObjectMethod("Derived", "Derived &opAssign(const Derived & in)", WRAP_OBJ_FIRST(assign<DerivedPtr>), asCALL_GENERIC); assert(r >= 0);

    r = engine->RegisterObjectMethod("Derived", "const int & get_x() const",  REF_GETTER_GEN(Derived, x), asCALL_GENERIC); assert(r >= 0);
    r = engine->RegisterObjectMethod("Derived", "void set_x(const int & in)", REF_SETTER_GEN(Derived, x), asCALL_GENERIC); assert(r >= 0);
    r = engine->RegisterObjectMethod("Derived", "int  get_y() const",             GETTER_GEN(Derived, y), asCALL_GENERIC); assert(r >= 0);
    r = engine->RegisterObjectMethod("Derived", "void set_y(int)",                SETTER_GEN(Derived, y), asCALL_GENERIC); assert(r >= 0);

    r = engine->RegisterObjectMethod("Derived", "void moo(void) const", GEN_CALLER(Derived, moo),    asCALL_GENERIC); assert(r >= 0);
    r = engine->RegisterObjectMethod("Derived", "void add(int)",        GEN_CALLER(Derived, add),    asCALL_GENERIC); assert(r >= 0);
    r = engine->RegisterObjectMethod("Derived", "int & getter(void)",   GEN_CALLER(Derived, getter), asCALL_GENERIC); assert(r >= 0);
  } else {
    r = engine->RegisterObjectBehaviour("Derived", asBEHAVE_CONSTRUCT,  "void f()",                   asFUNCTION(construct<DerivedPtr>),      asCALL_CDECL_OBJFIRST); assert(r >= 0);
    r = engine->RegisterObjectBehaviour("Derived", asBEHAVE_CONSTRUCT,  "void f(const Derived & in)", asFUNCTION(copy_construct<DerivedPtr>), asCALL_CDECL_OBJFIRST); assert(r >= 0);
    r = engine->RegisterObjectBehaviour("Derived", asBEHAVE_DESTRUCT,   "void f()",                   asFUNCTION(destroy<DerivedPtr>),        asCALL_CDECL_OBJFIRST); assert(r >= 0);
    r = engine->RegisterObjectMethod("Derived", "Derived &opAssign(const Derived & in)", asFUNCTION(assign<DerivedPtr>), asCALL_CDECL_OBJFIRST); assert(r >= 0);

    r = engine->RegisterObjectMethod("Derived", "const int & get_x() const",  REF_GETTER(Derived, x), asCALL_CDECL_OBJFIRST); assert(r >= 0);
    r = engine->RegisterObjectMethod("Derived", "void set_x(const int & in)", REF_SETTER(Derived, x), asCALL_CDECL_OBJFIRST); assert(r >= 0);
    r = engine->RegisterObjectMethod("Derived", "int  get_y() const",             GETTER(Derived, y), asCALL_CDECL_OBJFIRST); assert(r >= 0);
    r = engine->RegisterObjectMethod("Derived", "void set_y(int)",                SETTER(Derived, y), asCALL_CDECL_OBJFIRST); assert(r >= 0);

    r = engine->RegisterObjectMethod("Derived", "void moo(void) const", CALLER(Derived, moo),    asCALL_CDECL_OBJFIRST); assert(r >= 0);
    r = engine->RegisterObjectMethod("Derived", "void add(int)",        CALLER(Derived, add),    asCALL_CDECL_OBJFIRST); assert(r >= 0);
    r = engine->RegisterObjectMethod("Derived", "int & getter(void)",   CALLER(Derived, getter), asCALL_CDECL_OBJFIRST); assert(r >= 0);
  }

  // conversions
  if (use_generic) {
    r = engine->RegisterObjectBehaviour("Base", asBEHAVE_VALUE_CAST, "Derived f() const", WRAP_OBJ_FIRST((std::dynamic_pointer_cast<Derived, Base>)), asCALL_GENERIC); assert(r >= 0);
    r = engine->RegisterObjectBehaviour("Derived", asBEHAVE_IMPLICIT_VALUE_CAST, "Base f() const", WRAP_OBJ_FIRST((std::static_pointer_cast<Base, Derived>)), asCALL_GENERIC); assert(r >= 0);
  } else {
    r = engine->RegisterObjectBehaviour("Base", asBEHAVE_VALUE_CAST, "Derived f() const", asFUNCTION((std::dynamic_pointer_cast<Derived, Base>)), asCALL_CDECL_OBJFIRST); assert(r >= 0);
    r = engine->RegisterObjectBehaviour("Derived", asBEHAVE_IMPLICIT_VALUE_CAST, "Base f() const", asFUNCTION((std::static_pointer_cast<Base, Derived>)), asCALL_CDECL_OBJFIRST); assert(r >= 0);
  }

  const char script[] =
"void foo(const Base & in ptr) {\n"
"  ptr.moo();\n"
"  Derived d = Derived(ptr);\n"
"  print(d.x + \"\\n\");\n"
"  print(d.y + \"\\n\");\n"
"  d.getter() = 3;\n"
"  d.y = 4;\n"
"  d.moo();\n"
"  d.add(1);\n"
"}\n"
  ;
    
  asIScriptModule * mod = engine->GetModule(0, asGM_ALWAYS_CREATE);
  r = mod->AddScriptSection("script", script, sizeof(script) - 1); assert(r >= 0);
  r = mod->Build(); assert(r >= 0);
  
  int func_id = mod->GetFunctionIdByDecl("void foo(const Base & in)");
 
  Derived * d = new Derived;
  d->x = 1;
  d->y = 2;
  BasePtr base(d);

  
  asIScriptContext * ctx = engine->CreateContext();
  ctx->Prepare(func_id);
  ctx->SetArgObject(0, &base);
  ctx->Execute();
  
  ctx->Release();

  engine->Release();

  std::cout << d->x << std::endl;
  std::cout << d->y << std::endl;
}